/*
 * groupchatdlg.cpp - dialogs for handling groupchat
 * Copyright (C) 2001, 2002  Justin Karneges
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include"groupchatdlg.h"

#include<qlabel.h>
#include<qlayout.h>
#include<qpushbutton.h>
#include<qgroupbox.h>
#include<qmessagebox.h>
#include<qtextedit.h>
#include<qsplitter.h>
#include<qtimer.h>
#include<qheader.h>
#include<qtoolbutton.h>
#include<qinputdialog.h>
#include<qguardedptr.h>
#include<qaction.h>
#include<qobjectlist.h>
#include"psicon.h"
#include"psiaccount.h"
#include"busywidget.h"
#include"common.h"
#include"msgmle.h"

#ifdef Q_WS_WIN
#include<windows.h>
#endif


//----------------------------------------------------------------------------
// GCJoinDlg
//----------------------------------------------------------------------------
class GCJoinDlg::Private
{
public:
	Private() {}

	PsiCon *psi;
	PsiAccount *pa;
	AccountsComboBox *cb_ident;
	BusyWidget *busy;
	QStringList rl;
	Jid jid;
};

GCJoinDlg::GCJoinDlg(PsiCon *psi, PsiAccount *pa)
:GCJoinUI(0, 0, false, WDestructiveClose)
{
	d = new Private;
	d->psi = psi;
	d->pa = 0;
	d->psi->dialogRegister(this);
	updateIdentity(pa);

	d->cb_ident = d->psi->accountsComboBox(this);
	connect(d->cb_ident, SIGNAL(activated(PsiAccount *)), SLOT(updateIdentity(PsiAccount *)));
	d->cb_ident->setAccount(pa);
	replaceWidget(lb_ident, d->cb_ident);

	d->busy = new BusyWidget(this);
	replaceWidget(lb_busywidget, d->busy);

	d->rl = d->psi->recentGCList();
	for(QStringList::ConstIterator it = d->rl.begin(); it != d->rl.end(); ++it) {
		Jid j(*it);
		QString s = tr("%1 on %2").arg(j.resource()).arg(j.userHost());
		cb_recent->insertItem(s);
	}

	setCaption(CAP(tr("Join Groupchat")));
	pb_join->setDefault(true);
	connect(pb_close, SIGNAL(clicked()), SLOT(close()));
	connect(pb_join, SIGNAL(clicked()), SLOT(doJoin()));
	connect(cb_recent, SIGNAL(activated(int)), SLOT(recent_activated(int)));
	if(d->rl.isEmpty()) {
		cb_recent->setEnabled(false);
		le_host->setFocus();
	}
	else
		recent_activated(0);
}

GCJoinDlg::~GCJoinDlg()
{
	if(d->psi)
		d->psi->dialogUnregister(this);
	if(d->pa)
		d->pa->dialogUnregister(this);
	delete d;
}

void GCJoinDlg::closeEvent(QCloseEvent *e)
{
	e->ignore();
	reject();
}

void GCJoinDlg::done(int r)
{
	if(d->busy->isActive()) {
		int n = QMessageBox::information(this, tr("Warning"), tr("Are you sure you want to cancel joining groupchat?"), tr("&Yes"), tr("&No"));
		if(n != 0)
			return;
		d->pa->groupChatLeave(d->jid.host(), d->jid.user());
	}
	QDialog::done(r);
}

void GCJoinDlg::updateIdentity(PsiAccount *pa)
{
	if(d->pa)
		disconnect(d->pa, SIGNAL(disconnected()), this, SLOT(pa_disconnected()));

	d->pa = pa;

	if(!d->pa) {
		d->busy->stop();
		close();
		return;
	}

	connect(d->pa, SIGNAL(disconnected()), this, SLOT(pa_disconnected()));
}

void GCJoinDlg::pa_disconnected()
{
	if(d->busy->isActive()) {
		d->busy->stop();
		close();
	}
}

void GCJoinDlg::recent_activated(int x)
{
	int n = 0;
	bool found = false;
	QString str;
	for(QStringList::ConstIterator it = d->rl.begin(); it != d->rl.end(); ++it) {
		if(n == x) {
			found = true;
			str = *it;
			break;
		}
		++n;
	}
	if(!found)
		return;

	Jid j(str);
	le_host->setText(j.host());
	le_room->setText(j.user());
	le_nick->setText(j.resource());
}

void GCJoinDlg::doJoin()
{
	if(!d->pa->checkConnected(this))
		return;

	QString host = le_host->text();
	QString room = le_room->text();
	QString nick = le_nick->text();

	if(!d->pa->groupChatJoin(host, room, nick)) {
		QMessageBox::information(this, tr("Error"), tr("You are in or joining this room already!"));
		return;
	}

	if(host.isEmpty() || room.isEmpty() || nick.isEmpty()) {
		QMessageBox::information(this, tr("Error"), tr("You must fill out the fields in order to join."));
		return;
	}

	d->psi->dialogUnregister(this);
	d->jid = room + '@' + host + '/' + nick;
	d->pa->dialogRegister(this, d->jid);

	disableWidgets();
	d->busy->start();
}

void GCJoinDlg::disableWidgets()
{
	d->cb_ident->setEnabled(false);
	cb_recent->setEnabled(false);
	gb_info->setEnabled(false);
	pb_join->setEnabled(false);
}

void GCJoinDlg::enableWidgets()
{
	d->cb_ident->setEnabled(true);
	if(!d->rl.isEmpty())
		cb_recent->setEnabled(true);
	gb_info->setEnabled(true);
	pb_join->setEnabled(true);
}

void GCJoinDlg::joined()
{
	d->psi->recentGCAdd(d->jid.full());
	d->busy->stop();

	closeDialogs(this);
	deleteLater();
}

void GCJoinDlg::error(int, const QString &str)
{
	d->busy->stop();
	enableWidgets();

	pb_join->setFocus();

	d->pa->dialogUnregister(this);
	d->psi->dialogRegister(this);

	QMessageBox::information(this, tr("Error"), tr("Unable to join groupchat.\nReason: %1").arg(str));
}


//----------------------------------------------------------------------------
// GCUserView
//----------------------------------------------------------------------------
GCUserViewItem::GCUserViewItem(QListView *par)
:QListViewItem(par)
{
}

GCUserView::GCUserView(QWidget *parent, const char *name)
:QListView(parent, name)
{
	setResizeMode(QListView::AllColumns);
	setSorting(0);
	header()->hide();
	addColumn("");
}

GCUserView::~GCUserView()
{
}

void GCUserView::updateAll()
{
	for(GCUserViewItem *i = (GCUserViewItem *)firstChild(); i; i = (GCUserViewItem *)i->nextSibling())
		i->setPixmap(0, is.status(i->s));
}

QListViewItem *GCUserView::findEntry(const QString &nick)
{
	for(QListViewItem *lvi = firstChild(); lvi; lvi = lvi->nextSibling()) {
		if(lvi->text(0) == nick)
			return lvi;
	}
	return 0;
}

void GCUserView::updateEntry(const QString &nick, const Status &s)
{
	GCUserViewItem *lvi = (GCUserViewItem *)findEntry(nick);
	if(!lvi) {
		lvi = new GCUserViewItem(this);
		lvi->setText(0, nick);
	}

	lvi->s = s;
	lvi->setPixmap(0, is.status(lvi->s));
}

void GCUserView::removeEntry(const QString &nick)
{
	QListViewItem *lvi = findEntry(nick);
	if(lvi)
		delete lvi;
}


//----------------------------------------------------------------------------
// GCLineEdit
//----------------------------------------------------------------------------
GCLineEdit::GCLineEdit(QWidget *parent, const char *name)
:QLineEdit(parent, name)
{
}

void GCLineEdit::keyPressEvent(QKeyEvent *e)
{
	if(e->key() == Key_Up)
		upPressed();
	else if(e->key() == Key_Down)
		downPressed();

	QLineEdit::keyPressEvent(e);
}


//----------------------------------------------------------------------------
// GCMainDlg
//----------------------------------------------------------------------------
class GCMainDlg::Private
{
public:
	enum { Connecting, Connected, Idle };
	Private() {}

	int state;
	PsiAccount *pa;
	Jid jid;
	QString self;
	MsgMle *te_log;
	GCLineEdit *le_input;
	QLineEdit *le_topic;
	GCUserView *lv_users;
	QPushButton *pb_topic;
	QAction *act_find;
	int pending;

	QTimer *flashTimer;
	int flashCount;

	QStringList hist;
	int histAt;

	QGuardedPtr<GCFindDlg> findDlg;
	QString lastSearch;
};

GCMainDlg::GCMainDlg(PsiAccount *pa, const Jid &j)
:QWidget(0, 0, WDestructiveClose)
{
	d = new Private;
	d->pa = pa;
	d->jid = j.userHost();
	d->self = j.resource();
	d->pa->dialogRegister(this, d->jid);
	connect(d->pa, SIGNAL(updatedActivity()), SLOT(pa_updatedActivity()));

	d->pending = 0;
	d->flashTimer = 0;

	d->histAt = 0;
	d->findDlg = 0;

	d->state = Private::Connected;

	setIcon(is.groupchat);

	d->act_find = new QAction(tr("Find"), tr("&Find"), CTRL+Key_F, this);
	connect(d->act_find, SIGNAL(activated()), SLOT(openFind()));

	QVBoxLayout *vb = new QVBoxLayout(this, 4);
	QHBoxLayout *hb = new QHBoxLayout(vb);
	d->pb_topic = new QPushButton(tr("Topic:"), this);
	connect(d->pb_topic, SIGNAL(clicked()), SLOT(doTopic()));
	hb->addWidget(d->pb_topic);
	d->le_topic = new QLineEdit(this);
	d->le_topic->setReadOnly(true);
	hb->addWidget(d->le_topic);
	QToolButton *tb_find = new QToolButton(this);
	tb_find->setPixmap(*pix_find);
	connect(tb_find, SIGNAL(clicked()), SLOT(openFind()));
	hb->addWidget(tb_find);
	QLabel *lb_ident = d->pa->accountLabel(this, true);
	lb_ident->setSizePolicy(QSizePolicy( QSizePolicy::Minimum, QSizePolicy::Minimum ));
	hb->addWidget(lb_ident);

	QSplitter *sp = new QSplitter(this);
	vb->addWidget(sp);
	sp->setOrientation(QSplitter::Horizontal);
	d->te_log = new MsgMle(sp);
	d->te_log->setTextFormat(RichText);
	d->te_log->setReadOnly(true);
	d->te_log->setUndoRedoEnabled(false);
	d->te_log->setHScrollBarMode(QScrollView::AlwaysOff);
	d->te_log->setMimeSourceFactory(is.emoticons.generateFactory());

	d->lv_users = new GCUserView(sp);
	d->lv_users->setMinimumWidth(20);

	d->le_input = new GCLineEdit(this);
	vb->addWidget(d->le_input);
	connect(d->le_input, SIGNAL(returnPressed()), SLOT(le_returnPressed()));
	connect(d->le_input, SIGNAL(upPressed()), SLOT(le_upPressed()));
	connect(d->le_input, SIGNAL(downPressed()), SLOT(le_downPressed()));

	QValueList<int> list;
	list << 500;
	list << 80;
	sp->setSizes(list);

	resize(580,420);

	d->le_input->setFocus();

	setLooks();
	updateCaption();
}

GCMainDlg::~GCMainDlg()
{
	if(d->state != Private::Idle)
		d->pa->groupChatLeave(d->jid.host(), d->jid.user());

	QMimeSourceFactory *m = d->te_log->mimeSourceFactory();
	d->te_log->setMimeSourceFactory(0);
	delete m;

	d->pa->dialogUnregister(this);
	delete d;
}

void GCMainDlg::keyPressEvent(QKeyEvent *e)
{
	QWidget::keyPressEvent(e);
}

void GCMainDlg::closeEvent(QCloseEvent *e)
{
	e->accept();
}

void GCMainDlg::windowActivationChange(bool oldstate)
{
	QWidget::windowActivationChange(oldstate);

	// if we're bringing it to the front, get rid of the '*' if necessary
	if(isActiveWindow()) {
		if(d->pending > 0) {
			d->pending = 0;
			updateCaption();
		}
		doFlash(false);

		d->le_input->setFocus();
	}
}

void GCMainDlg::le_returnPressed()
{
	if(d->le_input->text().isEmpty())
		return;

	QString str = d->le_input->text();
	if(str == "/clear") {
		doClear();

		d->histAt = 0;
		d->hist.prepend(str);
		d->le_input->setText("");
		return;
	}

	if(d->state != Private::Connected)
		return;

	Message m(d->jid);
	m.setType("groupchat");
	m.setBody(str);
	m.setTimeStamp(QDateTime::currentDateTime());

	aSend(m);

	d->histAt = 0;
	d->hist.prepend(str);
	d->le_input->setText("");
}

void GCMainDlg::le_upPressed()
{
	if(d->histAt < (int)d->hist.count()) {
		++d->histAt;
		d->le_input->setText(d->hist[d->histAt-1]);
	}
}

void GCMainDlg::le_downPressed()
{
	if(d->histAt > 0) {
		--d->histAt;
		if(d->histAt == 0)
			d->le_input->setText("");
		else
			d->le_input->setText(d->hist[d->histAt-1]);
	}
}

void GCMainDlg::doTopic()
{
	bool ok = false;
	QString str = QInputDialog::getText(
		tr("Set Groupchat Topic"),
		tr("Enter a topic:"),
		QLineEdit::Normal, d->le_topic->text(), &ok, this);

	if(ok) {
		Message m(d->jid);
		m.setType("groupchat");
		m.setSubject(str);
		m.setBody(QString("/me ") + tr("has set the topic to: %1").arg(str));
		m.setTimeStamp(QDateTime::currentDateTime());
		aSend(m);
	}
}

void GCMainDlg::doClear()
{
	d->te_log->setText("");
}

void GCMainDlg::openFind()
{
	if(d->findDlg)
		bringToFront(d->findDlg);
	else {
		d->findDlg = new GCFindDlg(0, 0, d->lastSearch, this);
		connect(d->findDlg, SIGNAL(find(int, int, const QString &)), SLOT(doFind(int, int, const QString &)));
		d->findDlg->show();
	}
}

void GCMainDlg::doFind(int para, int index, const QString &str)
{
	d->lastSearch = str;

	if(!d->te_log->find(str, false, false, true, &para, &index))
		d->findDlg->error(str);
	else {
		// pass one character over
		d->findDlg->found(para, index+1);
	}
}

void GCMainDlg::goDisc()
{
	if(d->state != Private::Idle) {
		d->state = Private::Idle;
		d->pb_topic->setEnabled(false);
		appendSysMsg(tr("Disconnected."));
	}
}

void GCMainDlg::goConn()
{
	if(d->state == Private::Idle) {
		d->state = Private::Connecting;
		appendSysMsg(tr("Reconnecting..."));

		QString host = d->jid.host();
		QString room = d->jid.user();
		QString nick = d->self;

		if(!d->pa->groupChatJoin(host, room, nick)) {
			appendSysMsg(tr("Error: You are in or joining this room already!"));
			d->state = Private::Idle;
		}
	}
}

void GCMainDlg::pa_updatedActivity()
{
	if(!d->pa->loggedIn())
		goDisc();
	else {
		if(d->state == Private::Idle)
			goConn();
		else if(d->state == Private::Connected)
			d->pa->groupChatSetStatus(d->jid.host(), d->jid.user(), d->pa->status());
	}
}

void GCMainDlg::error(int, const QString &str)
{
	d->pb_topic->setEnabled(false);

	if(d->state == Private::Connecting)
		appendSysMsg(tr("Unable to join groupchat.  Reason: %1").arg(str));
	else
		appendSysMsg(tr("Unexpected groupchat error: %1").arg(str));

	d->state = Private::Idle;
}

void GCMainDlg::presence(const QString &nick, const Status &s)
{
	if(s.isAvailable())
		d->lv_users->updateEntry(nick, s);
	else
		d->lv_users->removeEntry(nick);
}

void GCMainDlg::message(const Message &m)
{
	QString from = m.from().resource();
	bool local = false;
	if(from == d->self)
		local = true;

	if(!m.subject().isEmpty()) {
		d->le_topic->setText(m.subject());
		d->le_topic->setCursorPosition(0);
		QToolTip::add(d->le_topic, QString("<qt><p>%1</p></qt>").arg(m.subject()));
	}

	if(from.isEmpty())
		appendSysMsg(m.body());
	else
		appendMessage(m, local);
}

void GCMainDlg::joined()
{
	if(d->state == Private::Connecting) {
		d->lv_users->clear();
		d->state = Private::Connected;
		d->pb_topic->setEnabled(true);
		appendSysMsg(tr("Connected."));
	}
}

void GCMainDlg::appendSysMsg(const QString &str)
{
	QString timestr;
	QDateTime time = QDateTime::currentDateTime();
	timestr.sprintf("%02d:%02d:%02d", time.time().hour(), time.time().minute(), time.time().second());

	int y = d->te_log->contentsHeight() - d->te_log->visibleHeight();
	if(y < 0)
		y = 0;
	bool atBottom = (d->te_log->contentsY() < y - 32) ? false: true;

	d->te_log->append(QString("<font color=\"#00A000\">[%1]").arg(timestr) + QString(" *** %1</font>").arg(str));

	if(atBottom)
		d->te_log->scrollToBottom();
}

void GCMainDlg::appendMessage(const Message &m, bool local)
{
	QString who, color;

	who = m.from().resource();
	if(local) {
		color = "#FF0000";
	}
	else {
		color = "#0000FF";
	}
	if(m.spooled())
		color = "#008000";

	QString timestr;
	QDateTime time = m.timeStamp();
	timestr.sprintf("%02d:%02d:%02d", time.time().hour(), time.time().minute(), time.time().second());

	int y = d->te_log->contentsHeight() - d->te_log->visibleHeight();
	if(y < 0)
		y = 0;
	bool atBottom = (d->te_log->contentsY() < y - 32) ? false: true;

	bool emote = false;
	if(m.body().left(4) == "/me ")
		emote = true;

	QString txt;
	if(emote)
		txt = plain2rich(m.body().mid(4));
	else
		txt = plain2rich(m.body());

	txt = linkify(txt);

	if(option.useEmoticons)
		txt = emoticonify(txt);

	if(emote) {
		d->te_log->append(QString("<font color=\"%1\">").arg(color) + QString("[%1]").arg(timestr) + QString(" *%1 ").arg(expandEntities(who)) + txt + "</font>");
	}
	else {
		if(option.chatSays)
			d->te_log->append(QString("<font color=\"%1\">").arg(color) + QString("[%1] ").arg(timestr) + QString("%1 says:").arg(expandEntities(who)) + "</font><br>" + txt);
		else
			d->te_log->append(QString("<font color=\"%1\">").arg(color) + QString("[%1] &lt;").arg(timestr) + expandEntities(who) + QString("&gt;</font> ") + txt);
	}

	if(local || atBottom)
		d->te_log->scrollToBottom();

	// if we're not active, notify the user by changing the title
	if(!isActiveWindow()) {
		++d->pending;
		updateCaption();
		doFlash(true);
	}

	/*if(!local) {
		d->keepOpen = true;
		QTimer::singleShot(1000, this, SLOT(setKeepOpenFalse()));
	}*/
}

void GCMainDlg::updateCaption()
{
	QString cap = "";

	if(d->pending > 0) {
		cap += "* ";
		if(d->pending > 1)
			cap += QString("[%1] ").arg(d->pending);
	}
	cap += d->jid.full();

	// if taskbar flash, then we need to erase first and redo
#ifdef Q_WS_WIN
	bool on = false;
	if(d->flashTimer)
		on = d->flashCount & 1;
	if(on)
		FlashWindow(winId(), true);
#endif
	setCaption(cap);
#ifdef Q_WS_WIN
	if(on)
		FlashWindow(winId(), true);
#endif
}

#ifdef Q_WS_WIN
void GCMainDlg::doFlash(bool yes)
{
	if(yes) {
		if(d->flashTimer)
			return;
		d->flashTimer = new QTimer(this);
		connect(d->flashTimer, SIGNAL(timeout()), SLOT(flashAnimate()));
		d->flashCount = 0;
		flashAnimate(); // kick the first one immediately
		d->flashTimer->start(500);
	}
	else {
		if(d->flashTimer) {
			delete d->flashTimer;
			d->flashTimer = 0;
			FlashWindow(winId(), false);
		}
	}
}
#else
void GCMainDlg::doFlash(bool)
{
}
#endif

void GCMainDlg::flashAnimate()
{
#ifdef Q_WS_WIN
	FlashWindow(winId(), true);
	++d->flashCount;
	if(d->flashCount == 5)
		d->flashTimer->stop();
#endif
}

void GCMainDlg::setLooks()
{
	// update the fonts
	QFont f;
	f.fromString(option.font[fChat]);
	d->te_log->setFont(f);
	d->le_input->setFont(f);

	f.fromString(option.font[fRoster]);
	d->lv_users->setFont(f);

	// update the widget icon
	setIcon(is.groupchat);
}

void GCMainDlg::optionsUpdate()
{
	QMimeSourceFactory *m = d->te_log->mimeSourceFactory();
	d->te_log->setMimeSourceFactory(is.emoticons.generateFactory());
	delete m;

	setLooks();

	// update status icons
	d->lv_users->updateAll();
}


//----------------------------------------------------------------------------
// GCFindDlg
//----------------------------------------------------------------------------
GCFindDlg::GCFindDlg(int startPara, int startIndex, const QString &str, QWidget *parent, const char *name)
:QDialog(parent, name, false, WDestructiveClose)
{
	para = startPara;
	index = startIndex;

	setCaption(tr("Find"));
	QVBoxLayout *vb = new QVBoxLayout(this, 4);
	QHBoxLayout *hb = new QHBoxLayout(vb);
	QLabel *l = new QLabel(tr("Find:"), this);
	hb->addWidget(l);
	le_input = new QLineEdit(this);
	hb->addWidget(le_input);
	vb->addStretch(1);

	QFrame *Line1 = new QFrame(this);
	Line1->setFrameShape( QFrame::HLine );
	Line1->setFrameShadow( QFrame::Sunken );
	Line1->setFrameShape( QFrame::HLine );
	vb->addWidget(Line1);

	hb = new QHBoxLayout(vb);
	hb->addStretch(1);
	QPushButton *pb_close = new QPushButton(tr("&Close"), this);
	connect(pb_close, SIGNAL(clicked()), SLOT(close()));
	hb->addWidget(pb_close);
	QPushButton *pb_find = new QPushButton(tr("&Find"), this);
	pb_find->setDefault(true);
	connect(pb_find, SIGNAL(clicked()), SLOT(doFind()));
	hb->addWidget(pb_find);
	pb_find->setAutoDefault(true);

	resize(200, minimumSize().height());

	le_input->setText(str);
	le_input->setFocus();
}

GCFindDlg::~GCFindDlg()
{
}

void GCFindDlg::found(int _para, int _index)
{
	para = _para;
	index = _index;
}

void GCFindDlg::error(const QString &str)
{
	QMessageBox::warning(this, tr("Find"), tr("Search string '%1' not found.").arg(str));
	le_input->setFocus();
}

void GCFindDlg::doFind()
{
	find(para, index, le_input->text());
}
